(function (global, factory) {
    typeof exports === 'object' && typeof module !== 'undefined' ? module.exports = factory() :
    typeof define === 'function' && define.amd ? define(factory) :
    (global = typeof globalThis !== 'undefined' ? globalThis : global || self, global.RouteRecognizer = factory());
}(this, (function () { 'use strict';

    const createObject = Object.create;
    function createMap() {
        const map = createObject(null);
        map["__"] = undefined;
        delete map["__"];
        return map;
    }

    class Target {
        constructor(path, matcher, delegate) {
            this.path = path;
            this.matcher = matcher;
            this.delegate = delegate;
        }
        to(target, callback) {
            const delegate = this.delegate;
            if (delegate && delegate.willAddRoute) {
                target = delegate.willAddRoute(this.matcher.target, target);
            }
            this.matcher.add(this.path, target);
            if (callback) {
                if (callback.length === 0) {
                    throw new Error("You must have an argument in the function passed to `to`");
                }
                this.matcher.addChild(this.path, target, callback, this.delegate);
            }
        }
    }
    class Matcher {
        constructor(target) {
            this.routes = createMap();
            this.children = createMap();
            this.target = target;
        }
        add(path, target) {
            this.routes[path] = target;
        }
        addChild(path, target, callback, delegate) {
            const matcher = new Matcher(target);
            this.children[path] = matcher;
            const match = generateMatch(path, matcher, delegate);
            if (delegate && delegate.contextEntered) {
                delegate.contextEntered(target, match);
            }
            callback(match);
        }
    }
    function generateMatch(startingPath, matcher, delegate) {
        function match(path, callback) {
            const fullPath = startingPath + path;
            if (callback) {
                callback(generateMatch(fullPath, matcher, delegate));
            }
            else {
                return new Target(fullPath, matcher, delegate);
            }
        }
        return match;
    }
    function addRoute(routeArray, path, handler) {
        let len = 0;
        for (let i = 0; i < routeArray.length; i++) {
            len += routeArray[i].path.length;
        }
        path = path.substr(len);
        const route = { path, handler };
        routeArray.push(route);
    }
    function eachRoute(baseRoute, matcher, callback, binding) {
        const routes = matcher.routes;
        const paths = Object.keys(routes);
        for (let i = 0; i < paths.length; i++) {
            const path = paths[i];
            const routeArray = baseRoute.slice();
            addRoute(routeArray, path, routes[path]);
            const nested = matcher.children[path];
            if (nested) {
                eachRoute(routeArray, nested, callback, binding);
            }
            else {
                callback.call(binding, routeArray);
            }
        }
    }
    function map(callback, addRouteCallback) {
        const matcher = new Matcher();
        callback(generateMatch("", matcher, this.delegate));
        eachRoute([], matcher, function (routes) {
            if (addRouteCallback) {
                addRouteCallback(this, routes);
            }
            else {
                this.add(routes);
            }
        }, this);
    }

    // Normalizes percent-encoded values in `path` to upper-case and decodes percent-encoded
    // values that are not reserved (i.e., unicode characters, emoji, etc). The reserved
    // chars are "/" and "%".
    // Safe to call multiple times on the same path.
    function normalizePath(path) {
        return path
            .split("/")
            .map(normalizeSegment)
            .join("/");
    }
    // We want to ensure the characters "%" and "/" remain in percent-encoded
    // form when normalizing paths, so replace them with their encoded form after
    // decoding the rest of the path
    const SEGMENT_RESERVED_CHARS = /%|\//g;
    function normalizeSegment(segment) {
        if (segment.length < 3 || segment.indexOf("%") === -1)
            return segment;
        return decodeURIComponent(segment).replace(SEGMENT_RESERVED_CHARS, encodeURIComponent);
    }
    // We do not want to encode these characters when generating dynamic path segments
    // See https://tools.ietf.org/html/rfc3986#section-3.3
    // sub-delims: "!", "$", "&", "'", "(", ")", "*", "+", ",", ";", "="
    // others allowed by RFC 3986: ":", "@"
    //
    // First encode the entire path segment, then decode any of the encoded special chars.
    //
    // The chars "!", "'", "(", ")", "*" do not get changed by `encodeURIComponent`,
    // so the possible encoded chars are:
    // ['%24', '%26', '%2B', '%2C', '%3B', '%3D', '%3A', '%40'].
    const PATH_SEGMENT_ENCODINGS = /%(?:2(?:4|6|B|C)|3(?:B|D|A)|40)/g;
    function encodePathSegment(str) {
        return encodeURIComponent(str).replace(PATH_SEGMENT_ENCODINGS, decodeURIComponent);
    }

    const escapeRegex = /(\/|\.|\*|\+|\?|\||\(|\)|\[|\]|\{|\}|\\)/g;
    const isArray = Array.isArray;
    // eslint-disable-next-line @typescript-eslint/unbound-method
    const hasOwnProperty = Object.prototype.hasOwnProperty;
    function getParam(params, key) {
        if (typeof params !== "object" || params === null) {
            throw new Error("You must pass an object as the second argument to `generate`.");
        }
        if (!hasOwnProperty.call(params, key)) {
            throw new Error("You must provide param `" + key + "` to `generate`.");
        }
        const value = params[key];
        const str = typeof value === "string" ? value : "" + value;
        if (str.length === 0) {
            throw new Error("You must provide a param `" + key + "`.");
        }
        return str;
    }
    const eachChar = [];
    eachChar[0 /* Static */] = function (segment, currentState) {
        let state = currentState;
        const value = segment.value;
        for (let i = 0; i < value.length; i++) {
            const ch = value.charCodeAt(i);
            state = state.put(ch, false, false);
        }
        return state;
    };
    eachChar[1 /* Dynamic */] = function (_, currentState) {
        return currentState.put(47 /* SLASH */, true, true);
    };
    eachChar[2 /* Star */] = function (_, currentState) {
        return currentState.put(-1 /* ANY */, false, true);
    };
    eachChar[4 /* Epsilon */] = function (_, currentState) {
        return currentState;
    };
    const regex = [];
    regex[0 /* Static */] = function (segment) {
        return segment.value.replace(escapeRegex, "\\$1");
    };
    regex[1 /* Dynamic */] = function () {
        return "([^/]+)";
    };
    regex[2 /* Star */] = function () {
        return "(.+)";
    };
    regex[4 /* Epsilon */] = function () {
        return "";
    };
    const generate = [];
    generate[0 /* Static */] = function (segment) {
        return segment.value;
    };
    generate[1 /* Dynamic */] = function (segment, params, shouldEncode) {
        const value = getParam(params, segment.value);
        if (shouldEncode) {
            return encodePathSegment(value);
        }
        else {
            return value;
        }
    };
    generate[2 /* Star */] = function (segment, params) {
        return getParam(params, segment.value);
    };
    generate[4 /* Epsilon */] = function () {
        return "";
    };
    const EmptyObject = Object.freeze({});
    const EmptyArray = Object.freeze([]);
    // The `names` will be populated with the paramter name for each dynamic/star
    // segment. `shouldDecodes` will be populated with a boolean for each dyanamic/star
    // segment, indicating whether it should be decoded during recognition.
    function parse(segments, route, types) {
        // normalize route as not starting with a "/". Recognition will
        // also normalize.
        if (route.length > 0 && route.charCodeAt(0) === 47 /* SLASH */) {
            route = route.substr(1);
        }
        const parts = route.split("/");
        let names = undefined;
        let shouldDecodes = undefined;
        for (let i = 0; i < parts.length; i++) {
            let part = parts[i];
            let type = 0;
            if (part === "") {
                type = 4 /* Epsilon */;
            }
            else if (part.charCodeAt(0) === 58 /* COLON */) {
                type = 1 /* Dynamic */;
            }
            else if (part.charCodeAt(0) === 42 /* STAR */) {
                type = 2 /* Star */;
            }
            else {
                type = 0 /* Static */;
            }
            if (type & 3 /* Named */) {
                part = part.slice(1);
                names = names || [];
                names.push(part);
                shouldDecodes = shouldDecodes || [];
                shouldDecodes.push((type & 1 /* Decoded */) !== 0);
            }
            if (type & 3 /* Counted */) {
                types[type]++;
            }
            segments.push({
                type,
                value: normalizeSegment(part)
            });
        }
        return {
            names: names || EmptyArray,
            shouldDecodes: shouldDecodes || EmptyArray
        };
    }
    function isEqualCharSpec(spec, char, negate) {
        return spec.char === char && spec.negate === negate;
    }
    // A State has a character specification and (`charSpec`) and a list of possible
    // subsequent states (`nextStates`).
    //
    // If a State is an accepting state, it will also have several additional
    // properties:
    //
    // * `regex`: A regular expression that is used to extract parameters from paths
    //   that reached this accepting state.
    // * `handlers`: Information on how to convert the list of captures into calls
    //   to registered handlers with the specified parameters
    // * `types`: How many static, dynamic or star segments in this route. Used to
    //   decide which route to use if multiple registered routes match a path.
    //
    // Currently, State is implemented naively by looping over `nextStates` and
    // comparing a character specification against a character. A more efficient
    // implementation would use a hash of keys pointing at one or more next states.
    class State {
        constructor(states, id, char, negate, repeat) {
            this.states = states;
            this.id = id;
            this.char = char;
            this.negate = negate;
            this.nextStates = repeat ? id : null;
            this.pattern = "";
            this._regex = undefined;
            this.handlers = undefined;
            this.types = undefined;
        }
        regex() {
            if (!this._regex) {
                this._regex = new RegExp(this.pattern);
            }
            return this._regex;
        }
        get(char, negate) {
            const nextStates = this.nextStates;
            if (nextStates === null)
                return;
            if (isArray(nextStates)) {
                for (let i = 0; i < nextStates.length; i++) {
                    const child = this.states[nextStates[i]];
                    if (isEqualCharSpec(child, char, negate)) {
                        return child;
                    }
                }
            }
            else {
                const child = this.states[nextStates];
                if (isEqualCharSpec(child, char, negate)) {
                    return child;
                }
            }
        }
        put(char, negate, repeat) {
            let state;
            // If the character specification already exists in a child of the current
            // state, just return that state.
            if ((state = this.get(char, negate))) {
                return state;
            }
            // Make a new state for the character spec
            const states = this.states;
            state = new State(states, states.length, char, negate, repeat);
            states[states.length] = state;
            // Insert the new state as a child of the current state
            if (this.nextStates == null) {
                this.nextStates = state.id;
            }
            else if (isArray(this.nextStates)) {
                this.nextStates.push(state.id);
            }
            else {
                this.nextStates = [this.nextStates, state.id];
            }
            // Return the new state
            return state;
        }
        // Find a list of child states matching the next character
        match(ch) {
            const nextStates = this.nextStates;
            if (!nextStates)
                return [];
            const returned = [];
            if (isArray(nextStates)) {
                for (let i = 0; i < nextStates.length; i++) {
                    const child = this.states[nextStates[i]];
                    if (isMatch(child, ch)) {
                        returned.push(child);
                    }
                }
            }
            else {
                const child = this.states[nextStates];
                if (isMatch(child, ch)) {
                    returned.push(child);
                }
            }
            return returned;
        }
    }
    function isMatch(spec, char) {
        return spec.negate
            ? spec.char !== char && spec.char !== -1 /* ANY */
            : spec.char === char || spec.char === -1 /* ANY */;
    }
    // This is a somewhat naive strategy, but should work in a lot of cases
    // A better strategy would properly resolve /posts/:id/new and /posts/edit/:id.
    //
    // This strategy generally prefers more static and less dynamic matching.
    // Specifically, it
    //
    //  * prefers fewer stars to more, then
    //  * prefers using stars for less of the match to more, then
    //  * prefers fewer dynamic segments to more, then
    //  * prefers more static segments to more
    function sortSolutions(states) {
        return states.sort(function (a, b) {
            const [astatics, adynamics, astars] = a.types || [0, 0, 0];
            const [bstatics, bdynamics, bstars] = b.types || [0, 0, 0];
            if (astars !== bstars) {
                return astars - bstars;
            }
            if (astars) {
                if (astatics !== bstatics) {
                    return bstatics - astatics;
                }
                if (adynamics !== bdynamics) {
                    return bdynamics - adynamics;
                }
            }
            if (adynamics !== bdynamics) {
                return adynamics - bdynamics;
            }
            if (astatics !== bstatics) {
                return bstatics - astatics;
            }
            return 0;
        });
    }
    function recognizeChar(states, ch) {
        let nextStates = [];
        for (let i = 0, l = states.length; i < l; i++) {
            const state = states[i];
            nextStates = nextStates.concat(state.match(ch));
        }
        return nextStates;
    }
    class RecognizeResults {
        constructor(queryParams) {
            this.length = 0;
            this.queryParams = queryParams || {};
        }
    }
    // eslint-disable-next-line @typescript-eslint/unbound-method
    RecognizeResults.prototype.splice = Array.prototype.splice;
    // eslint-disable-next-line @typescript-eslint/unbound-method
    RecognizeResults.prototype.slice = Array.prototype.slice;
    // eslint-disable-next-line @typescript-eslint/unbound-method
    RecognizeResults.prototype.push = Array.prototype.push;
    function findHandler(state, originalPath, queryParams, shouldDecode) {
        const handlers = state.handlers;
        const regex = state.regex();
        if (!regex || !handlers)
            throw new Error("state not initialized");
        const captures = regex.exec(originalPath);
        let currentCapture = 1;
        const result = new RecognizeResults(queryParams);
        result.length = handlers.length;
        for (let i = 0; i < handlers.length; i++) {
            const handler = handlers[i];
            const names = handler.names;
            const shouldDecodes = handler.shouldDecodes;
            let params = EmptyObject;
            let isDynamic = false;
            if (names !== EmptyArray && shouldDecodes !== EmptyArray) {
                for (let j = 0; j < names.length; j++) {
                    isDynamic = true;
                    const name = names[j];
                    const capture = captures && captures[currentCapture++];
                    if (params === EmptyObject) {
                        params = {};
                    }
                    if (shouldDecode && shouldDecodes[j]) {
                        params[name] = capture && decodeURIComponent(capture);
                    }
                    else {
                        params[name] = capture;
                    }
                }
            }
            result[i] = {
                handler: handler.handler,
                params,
                isDynamic
            };
        }
        return result;
    }
    function decodeQueryParamPart(part) {
        // http://www.w3.org/TR/html401/interact/forms.html#h-17.13.4.1
        part = part.replace(/\+/gm, "%20");
        let result;
        try {
            result = decodeURIComponent(part);
        }
        catch (error) {
            result = "";
        }
        return result;
    }
    class RouteRecognizer {
        constructor() {
            this.names = createMap();
            const states = [];
            const state = new State(states, 0, -1 /* ANY */, true, false);
            states[0] = state;
            this.rootState = state;
        }
        add(routes, options) {
            let currentState = this.rootState;
            let pattern = "^";
            const types = [0, 0, 0];
            const handlers = new Array(routes.length);
            const allSegments = [];
            let isEmpty = true;
            let j = 0;
            for (let i = 0; i < routes.length; i++) {
                const route = routes[i];
                const { names, shouldDecodes } = parse(allSegments, route.path, types);
                // preserve j so it points to the start of newly added segments
                for (; j < allSegments.length; j++) {
                    const segment = allSegments[j];
                    if (segment.type === 4 /* Epsilon */) {
                        continue;
                    }
                    isEmpty = false;
                    // Add a "/" for the new segment
                    currentState = currentState.put(47 /* SLASH */, false, false);
                    pattern += "/";
                    // Add a representation of the segment to the NFA and regex
                    currentState = eachChar[segment.type](segment, currentState);
                    pattern += regex[segment.type](segment);
                }
                handlers[i] = {
                    handler: route.handler,
                    names,
                    shouldDecodes
                };
            }
            if (isEmpty) {
                currentState = currentState.put(47 /* SLASH */, false, false);
                pattern += "/";
            }
            currentState.handlers = handlers;
            currentState.pattern = pattern + "$";
            currentState.types = types;
            let name;
            if (typeof options === "object" && options !== null && options.as) {
                name = options.as;
            }
            if (name) {
                // if (this.names[name]) {
                //   throw new Error("You may not add a duplicate route named `" + name + "`.");
                // }
                this.names[name] = {
                    segments: allSegments,
                    handlers
                };
            }
        }
        handlersFor(name) {
            const route = this.names[name];
            if (!route) {
                throw new Error("There is no route named " + name);
            }
            const result = new Array(route.handlers.length);
            for (let i = 0; i < route.handlers.length; i++) {
                const handler = route.handlers[i];
                result[i] = handler;
            }
            return result;
        }
        hasRoute(name) {
            return !!this.names[name];
        }
        generate(name, params) {
            const shouldEncode = RouteRecognizer.ENCODE_AND_DECODE_PATH_SEGMENTS;
            const route = this.names[name];
            let output = "";
            if (!route) {
                throw new Error("There is no route named " + name);
            }
            const segments = route.segments;
            for (let i = 0; i < segments.length; i++) {
                const segment = segments[i];
                if (segment.type === 4 /* Epsilon */) {
                    continue;
                }
                output += "/";
                output += generate[segment.type](segment, params, shouldEncode);
            }
            if (!output.startsWith("/")) {
                output = "/" + output;
            }
            if (params && params.queryParams) {
                output += this.generateQueryString(params.queryParams);
            }
            return output;
        }
        generateQueryString(params) {
            const pairs = [];
            const keys = Object.keys(params);
            keys.sort();
            for (let i = 0; i < keys.length; i++) {
                const key = keys[i];
                const value = params[key];
                if (value == null) {
                    continue;
                }
                let pair = encodeURIComponent(key);
                if (isArray(value)) {
                    for (let j = 0; j < value.length; j++) {
                        const arrayPair = key + "[]" + "=" + encodeURIComponent(value[j]);
                        pairs.push(arrayPair);
                    }
                }
                else {
                    pair += "=" + encodeURIComponent(value);
                    pairs.push(pair);
                }
            }
            if (pairs.length === 0) {
                return "";
            }
            return "?" + pairs.join("&");
        }
        parseQueryString(queryString) {
            const pairs = queryString.split("&");
            const queryParams = {};
            for (let i = 0; i < pairs.length; i++) {
                const pair = pairs[i].split("=");
                let key = decodeQueryParamPart(pair[0]);
                const keyLength = key.length;
                let isArray = false;
                let value;
                if (pair.length === 1) {
                    value = "true";
                }
                else {
                    // Handle arrays
                    if (keyLength > 2 && key.endsWith("[]")) {
                        isArray = true;
                        key = key.slice(0, keyLength - 2);
                        if (!queryParams[key]) {
                            queryParams[key] = [];
                        }
                    }
                    value = pair[1] ? decodeQueryParamPart(pair[1]) : "";
                }
                if (isArray) {
                    queryParams[key].push(value);
                }
                else {
                    queryParams[key] = value;
                }
            }
            return queryParams;
        }
        recognize(path) {
            const shouldNormalize = RouteRecognizer.ENCODE_AND_DECODE_PATH_SEGMENTS;
            let results;
            let states = [this.rootState];
            let queryParams = {};
            let isSlashDropped = false;
            const hashStart = path.indexOf("#");
            if (hashStart !== -1) {
                path = path.substr(0, hashStart);
            }
            const queryStart = path.indexOf("?");
            if (queryStart !== -1) {
                const queryString = path.substr(queryStart + 1, path.length);
                path = path.substr(0, queryStart);
                queryParams = this.parseQueryString(queryString);
            }
            if (!path.startsWith("/")) {
                path = "/" + path;
            }
            let originalPath = path;
            if (shouldNormalize) {
                path = normalizePath(path);
            }
            else {
                path = decodeURI(path);
                originalPath = decodeURI(originalPath);
            }
            const pathLen = path.length;
            if (pathLen > 1 && path.charAt(pathLen - 1) === "/") {
                path = path.substr(0, pathLen - 1);
                originalPath = originalPath.substr(0, originalPath.length - 1);
                isSlashDropped = true;
            }
            for (let i = 0; i < path.length; i++) {
                states = recognizeChar(states, path.charCodeAt(i));
                if (!states.length) {
                    break;
                }
            }
            const solutions = [];
            for (let i = 0; i < states.length; i++) {
                if (states[i].handlers) {
                    solutions.push(states[i]);
                }
            }
            states = sortSolutions(solutions);
            const state = solutions[0];
            if (state && state.handlers) {
                // if a trailing slash was dropped and a star segment is the last segment
                // specified, put the trailing slash back
                if (isSlashDropped && state.char === -1 /* ANY */) {
                    originalPath = originalPath + "/";
                }
                results = findHandler(state, originalPath, queryParams, shouldNormalize);
            }
            return results;
        }
    }
    RouteRecognizer.VERSION = "VERSION_STRING_PLACEHOLDER";
    // Set to false to opt-out of encoding and decoding path segments.
    // See https://github.com/tildeio/route-recognizer/pull/55
    RouteRecognizer.ENCODE_AND_DECODE_PATH_SEGMENTS = true;
    RouteRecognizer.Normalizer = {
        normalizeSegment,
        normalizePath,
        encodePathSegment
    };
    RouteRecognizer.prototype.map = map;

    return RouteRecognizer;

})));
